@@ -105,7 +105,14 @@ class ABI_EXPORT GR_Font
 	   NB: it is essential that this function is fast
 	*/
 	virtual bool doesGlyphExist(UT_UCS4Char g);
+//
+// UT_Rect of glyph in Logical units.
+// rec.left = bearing Left (distance from origin to start)
+// rec.width = width of the glyph
+// rec.top = distance from the origin to the top of the glyph
+// rec.height = total height of the glyph
 
+	virtual bool glyphBox(UT_UCS4Char g, UT_Rect & rec) const = 0;
 	static  bool s_doesGlyphExist(UT_UCS4Char g, void *instance)
 	{
 		UT_return_val_if_fail(instance, false);
@@ -167,7 +174,7 @@ enum GR_GraphicsId
 
 	/* IDs for built-in classes: DO NOT CHANGE THE ASSIGNED VALUES !!!*/
 	/* (these classes cannot be unregistered) */
-	GRID_BEOS            =  0x101,
+	GRID_BEOS            =  0x101, // deprecated
 	GRID_COCOA           =  0x102,
 	GRID_QNX             =  0x103,
 	GRID_WIN32           =  0x104,
@@ -247,6 +254,8 @@ enum GR_Capability
 class GR_AllocInfo
 {
   public:
+	virtual ~GR_AllocInfo() {}
+
 	virtual GR_GraphicsId getType() const {UT_ASSERT_HARMLESS(UT_NOT_IMPLEMENTED); return GRID_UNKNOWN;}
 	virtual bool isPrinterGraphics()const {UT_ASSERT_HARMLESS(UT_NOT_IMPLEMENTED); return false;}
 };
@@ -373,12 +382,14 @@ class ABI_EXPORT GR_Graphics
 					   const char* pszFontWeight,
 					   const char* pszFontStretch,
 					   const char* pszFontSize);
-	static const char* findNearestFont(const char* pszFontFamily,
-									   const char* pszFontStyle,
-									   const char* pszFontVariant,
-									   const char* pszFontWeight,
-									   const char* pszFontStretch,
-									   const char* pszFontSize);
+	
+	virtual const char* findNearestFont(const char* pszFontFamily,
+										const char* pszFontStyle,
+										const char* pszFontVariant,
+										const char* pszFontWeight,
+										const char* pszFontStretch,
+										const char* pszFontSize)
+	                                         {return pszFontFamily;}
 
 	const char *      invertDimension(UT_Dimension, double) const;
 
@@ -485,7 +496,9 @@ class ABI_EXPORT GR_Graphics
 				   GR_CURSOR_VLINE_DRAG,
 				   GR_CURSOR_HLINE_DRAG,
 				   GR_CURSOR_CROSSHAIR,
-				   GR_CURSOR_DOWNARROW
+		                   GR_CURSOR_DOWNARROW,
+		                   GR_CURSOR_DRAGTEXT,
+		                   GR_CURSOR_COPYTEXT
 	} Cursor;
 
 	virtual void      setCursor(GR_Graphics::Cursor c) = 0;
@@ -506,6 +519,7 @@ class ABI_EXPORT GR_Graphics
 #define COUNT_3D_COLORS 5
 
 	virtual void      setColor3D(GR_Color3D c) = 0;
+	virtual bool      getColor3D(GR_Color3D name, UT_RGBColor &color) { return false; }
 
 	//
 	// Methods to deal with background repainting as used in the Unix FE. These
@@ -738,6 +752,8 @@ class ABI_EXPORT GR_Graphics
 
 	const UT_Rect *  m_pRect;
 
+	bool m_bHave3DColors;
+
  private:
 	GR_Caret *		 m_pCaret;
     bool             _PtInPolygon(UT_Point * pts,UT_uint32 nPoints,UT_sint32 x,UT_sint32 y);
