@@ -35,7 +35,7 @@
 #include "xap_UnixFontXLFD.h"
 
 #include "ut_AdobeEncoding.h"
-
+#include "ut_string_class.h"
 #include <ft2build.h>
 #include FT_FREETYPE_H
 #include <X11/Xft/Xft.h>
@@ -68,6 +68,10 @@ class ABI_EXPORT XAP_UnixFont
 {
  public:
 
+	/*! 
+	* \todo ROB: we maybe could move to using pango attributes exclusively here 
+	* since this is unix only.
+	*/
 	enum style
 	{
 		STYLE_NORMAL = 0,
@@ -96,6 +100,9 @@ class ABI_EXPORT XAP_UnixFont
 	void					setStyle(XAP_UnixFont::style s);
 	XAP_UnixFont::style		getStyle(void) const;
 
+	PangoStyle				getPangoStyle(void) const;
+	PangoWeight				getPangoWeight(void) const;
+
 	const char * 			getFontfile(void) const;
 	const char * 			getMetricfile(void) const;
 	const encoding_pair*	loadEncodingFile(void);
@@ -105,7 +112,8 @@ class ABI_EXPORT XAP_UnixFont
 
 	void					setXLFD(const char * xlfd);
 	const char * 			getXLFD(void) const;
-	
+	bool                    isSymbol(void)
+		{ return m_bIsSymbol;}
 	bool					embedInto(ps_Generate& ps);
 	bool					openPFA(void);
 	char					getPFAChar(void);
@@ -118,7 +126,7 @@ class ABI_EXPORT XAP_UnixFont
 	bool					is_PS_font()  const {return ((m_fontType == FONT_TYPE_PFA) || (m_fontType == FONT_TYPE_PFB));}
 	font_type				getFontType() const {return m_fontType;}
 
-	void					getCoverage(UT_Vector& coverage);
+	void					getCoverage(UT_NumberVector& coverage);
 
 	float					getAscender(UT_uint32 iSize) const;
 	float					getDescender(UT_uint32 iSize) const;
@@ -127,8 +135,14 @@ class ABI_EXPORT XAP_UnixFont
 	void					setFontManager(XAP_UnixFontManager * pFm);
 	
 	bool doesGlyphExist(UT_UCS4Char g);
+	bool                    isSymbol(void) const;
+	bool                    isDingbat(void) const;
 
 private:
+
+	XAP_UnixFont ();
+	XAP_UnixFont & operator= (const XAP_UnixFont & rhs);
+
 	bool					_createTtfSupportFiles();
 	bool					_createPsSupportFiles();
 	struct allocFont
@@ -141,11 +155,11 @@ private:
 	char * 					m_fontKey;
 
 	// a cache of XftFont * at a given size
-	mutable UT_Vector		m_allocFonts;
+	mutable UT_GenericVector<allocFont *>	m_allocFonts;
 	
 	char * 					m_name;
 	XAP_UnixFont::style		m_style;
-	char * 					m_xlfd;
+	UT_String    			m_xlfd;
 	
 	// The next line is the info that is given back to us by parseAFM. The line after that is our own mangled one to follow Unicode.
 	uniWidth *				m_uniWidths;
@@ -200,6 +214,8 @@ class XAP_UnixFontHandle : public GR_Font
 
 	XAP_UnixFontHandle(XAP_UnixFont * font, UT_uint32 size);	
  
+	virtual ~XAP_UnixFontHandle () {}
+
 	XftFont * 					getLayoutXftFont(void);
 	XftFont * 					getDeviceXftFont(UT_uint32 zoomPercentage);
 
@@ -214,9 +230,13 @@ class XAP_UnixFontHandle : public GR_Font
 	float measureUnRemappedChar(const UT_UCSChar c, UT_uint32 iSize) const;
 
 	virtual bool doesGlyphExist(UT_UCS4Char g);
+        virtual bool            glyphBox(UT_UCS4Char g, UT_Rect & rec) const;
 
 private:
  	XAP_UnixFontHandle();
+ 	XAP_UnixFontHandle(const XAP_UnixFontHandle & rhs);
+ 	XAP_UnixFontHandle & operator=(const XAP_UnixFontHandle & rhs);
+
 	XAP_UnixFont*				m_font;
 	UT_uint32					m_size;
 };
