@@ -1,3 +1,5 @@
+/* -*- mode: C++; tab-width: 4; c-basic-offset: 4; -*- */
+
 /* AbiSource Application Framework
  * Copyright (C) 2000
  * Orignially by Vlad Harchev <hvv@hippo.ru>
@@ -29,10 +31,11 @@
 #include <stdio.h>
 #include <string.h>
 
-
-static 	UT_iconv_t iconv_handle_N2U = NULL, iconv_handle_U2N = NULL,
-	iconv_handle_U2Latin1 = NULL,
-	iconv_handle_U2Win = NULL ,iconv_handle_Win2U = NULL;
+static UT_iconv_t iconv_handle_N2U      = UT_ICONV_INVALID;
+static UT_iconv_t iconv_handle_U2N      = UT_ICONV_INVALID;
+static UT_iconv_t iconv_handle_U2Latin1 = UT_ICONV_INVALID;
+static UT_iconv_t iconv_handle_U2Win    = UT_ICONV_INVALID;
+static UT_iconv_t iconv_handle_Win2U    = UT_ICONV_INVALID;
 
 XAP_EncodingManager*	XAP_EncodingManager::_instance = NULL;
 
@@ -71,6 +74,17 @@ const char* XAP_EncodingManager::getNative8BitEncodingName() const
     return getNativeEncodingName();
 }
 
+/*!
+ * Returns the native non-Unicode encoding
+ *
+ * Always returns a non-Unicode encoding, even when the native
+ * encoding is Unicode like UCS-2 on Windows NT or UTF-8 on *nix.
+ */
+const char* XAP_EncodingManager::getNativeNonUnicodeEncodingName() const
+{
+    return getNativeEncodingName();
+}
+
 /*!
  * Returns the native Unicode encoding
  *
@@ -81,7 +95,10 @@ const char* XAP_EncodingManager::getNativeUnicodeEncodingName() const
     return "UTF-8"; /* this will definitely work*/
 }
 
-static const char* UCS2BEName, *UCS2LEName;
+static const char * UCS2BEName = 0;
+static const char * UCS2LEName = 0;
+static const char * UCS4BEName = 0;
+static const char * UCS4LEName = 0;
 
 /*!
  * Returns the name this system uses for UCS-2, big endian
@@ -103,23 +120,33 @@ const char* XAP_EncodingManager::getUCS2LEName() const
 	return UCS2LEName;
 }
 
+/*!
+ * Returns the name this system uses for UCS-4, big endian
+ *
+ * UCS-4BE is standard
+ */
+const char* XAP_EncodingManager::getUCS4BEName() const
+{
+	return UCS4BEName;
+}
+
+/*!
+ * Returns the name this system uses for UCS-4, little endian
+ *
+ * UCS-4LE is standard
+ */
+const char* XAP_EncodingManager::getUCS4LEName() const
+{
+	return UCS4LEName;
+}
+
 XAP_EncodingManager::~XAP_EncodingManager()
 {
-	UT_DEBUGMSG(("CLOSING XAP_ENCODINGMANAGER\n"));
-	if(UT_iconv_isValid(iconv_handle_N2U))
-		UT_iconv_close(iconv_handle_N2U);
-
-	if(UT_iconv_isValid(iconv_handle_U2N))
-		UT_iconv_close(iconv_handle_U2N);
-
-	if(UT_iconv_isValid(iconv_handle_U2Latin1))
-		UT_iconv_close(iconv_handle_U2Latin1);
-
-	if(UT_iconv_isValid(iconv_handle_U2Win))
-		UT_iconv_close(iconv_handle_U2Win);
-  
-	if(UT_iconv_isValid(iconv_handle_Win2U))
-		UT_iconv_close(iconv_handle_Win2U);
+	UT_iconv_close(iconv_handle_N2U);
+	UT_iconv_close(iconv_handle_U2N);
+	UT_iconv_close(iconv_handle_U2Latin1);
+	UT_iconv_close(iconv_handle_U2Win);
+	UT_iconv_close(iconv_handle_Win2U);
 }
 
 XAP_EncodingManager::XAP_EncodingManager() { }
@@ -133,12 +160,12 @@ void XAP_EncodingManager::Delete_instance()
 const char* XAP_EncodingManager::getLanguageISOName() const 
 {
     return "en";
-};
+}
 
 const char* XAP_EncodingManager::getLanguageISOTerritory() const
 {
     return NULL;
-};
+}
 
 // TODO Do we need an equivalent function which can return
 // TODO U+FFFD "REPLACEMENT CHARACTER" or U+25A0 "BLACK SQUARE"
@@ -173,13 +200,13 @@ UT_uint32 XAP_EncodingManager::approximate(char* out,UT_uint32 max_length,UT_UCS
 		*/
 	}
 	return 0;
-};
+}
 
 UT_UCSChar XAP_EncodingManager::nativeToU(UT_UCSChar c) const
 {
     UT_UCSChar ret = try_nativeToU(c);
     return ret ? ret : fallbackChar(c);	
-};
+}
 
 UT_UCSChar XAP_EncodingManager::UToNative(UT_UCSChar c)  const
 {
@@ -192,35 +219,35 @@ UT_UCSChar XAP_EncodingManager::UToNative(UT_UCSChar c)  const
     }
     else
     	return ret;
-};
+}
 
 UT_UCSChar XAP_EncodingManager::WindowsToU(UT_UCSChar c) const
 {
     UT_UCSChar ret = try_WindowsToU(c);
     return ret ? ret : fallbackChar(c);	
-};
+}
 
 
 UT_UCSChar XAP_EncodingManager::UToWindows(UT_UCSChar c)  const
 {
     UT_UCSChar ret = try_UToWindows(c);
     return ret && ret<=0xff ? ret : fallbackChar(c);	
-};
+}
 
 
 const char* XAP_EncodingManager::strToNative(const char* in,const char* charset,bool bUseSysEncoding) const
 {
 	static char buf[500];
 	return strToNative(in, charset, buf, sizeof(buf), bUseSysEncoding);
-};
+}
 
 const char* XAP_EncodingManager::strToNative(const char* in, const char* charset, char* buf, int bufsz,
 											 bool bUseSysEncoding) const
 {
 	if (!charset || !*charset || !in || !*in || !buf)
 		return in; /*won't translate*/
-#if 0
-	// TODO this gets around the fact that gtk 1.2 cannot handle utf-8 input
+#if 1
+	// TODO this gets around the fact that gtk 1.2 cannot handle UTF-8 input
 	// when we move to gtk 2 this original branch needs to be enabled
 	UT_iconv_t iconv_handle = UT_iconv_open(
 		bUseSysEncoding ? getNativeSystemEncodingName() : getNativeEncodingName(), charset);
@@ -228,8 +255,8 @@ const char* XAP_EncodingManager::strToNative(const char* in, const char* charset
 	UT_iconv_t iconv_handle;
 	const char * pNative =  bUseSysEncoding ? getNativeSystemEncodingName() : getNativeEncodingName();
 	
-	if(!strcmp(pNative, "utf-8") || !strcmp(pNative, "UTF-8"))
-		pNative = getNative8BitEncodingName();
+	if(!UT_stricmp(pNative, "UTF-8"))
+		pNative = getNativeNonUnicodeEncodingName();
 
 	iconv_handle = UT_iconv_open(pNative, charset);
 	xxx_UT_DEBUGMSG(("xap_EncodingManager::strToNative: pNative %s, iconv_handle 0x%x\n",pNative, iconv_handle));
@@ -253,13 +280,13 @@ const char* XAP_EncodingManager::strToNative(const char* in, const char* charset
 
 	UT_iconv_close(iconv_handle);
 	return retstr;
-};
+}
 
 int XAP_EncodingManager::XAP_XML_UnknownEncodingHandler(void* /*encodingHandlerData*/,
                                           const XML_Char *name,
                                           XML_Encoding *info)
 {
-#ifdef	HAVE_LIBXML2
+#ifndef	HAVE_EXPAT
 	return 0;
 #else
 	/*this is used by code that reads xml using expat*/
@@ -277,15 +304,15 @@ int XAP_EncodingManager::XAP_XML_UnknownEncodingHandler(void* /*encodingHandlerD
 			size_t ibuflen = 1, obuflen=2;
 			const char* iptr = ibuf;
 			char* optr = obuf;
-			ibuf[0] = (unsigned char)i;
+			ibuf[0] = static_cast<unsigned char>(i);
 			size_t donecnt = UT_iconv(iconv_handle,&iptr,&ibuflen,&optr,&obuflen);			
 			if (donecnt!=(size_t)-1 && ibuflen==0) 
 			{
 				unsigned short uval;
-				unsigned short b0 = (unsigned char)obuf[swap_stou];
-				unsigned short b1 = (unsigned char)obuf[!swap_stou];
+				unsigned short b0 = static_cast<unsigned char>(obuf[swap_stou]);
+				unsigned short b1 = static_cast<unsigned char>(obuf[!swap_stou]);
 				uval =  b0 | (b1<<8);
-				info->map[i] = (unsigned int) uval;
+				info->map[i] = static_cast<unsigned int>(uval);
 			}
 			else
 				info->map[i] = -1;/* malformed character. Such cases exist - e.g. 0x98 in cp1251*/
@@ -295,88 +322,141 @@ int XAP_EncodingManager::XAP_XML_UnknownEncodingHandler(void* /*encodingHandlerD
 	UT_iconv_close(iconv_handle);
 	return 1;
 #endif
-};
+}
 
 extern "C" { char *wvLIDToCodePageConverter(unsigned short lid); }
+
 static void init_values(const XAP_EncodingManager* that)
 {
-	iconv_handle_N2U = UT_iconv_open("UCS-2",that->getNativeEncodingName());
-	iconv_handle_U2N = UT_iconv_open(that->getNativeEncodingName(),"UCS-2");
-	iconv_handle_U2Latin1 = UT_iconv_open("ISO-8859-1","UCS-2");
+	const char * ucs4i = ucs4Internal ();
+	const char * naten = that->getNativeEncodingName ();
+
+	iconv_handle_N2U = UT_iconv_open (ucs4i, naten);
+	if (!UT_iconv_isValid(iconv_handle_N2U))
+		{
+			UT_DEBUGMSG(("WARNING: UT_iconv_open(%s,%s) failed!\n",ucs4i,naten));
+		}
+	iconv_handle_U2N = UT_iconv_open (naten, ucs4i);
+	if (!UT_iconv_isValid(iconv_handle_U2N))
+		{
+			UT_DEBUGMSG(("WARNING: UT_iconv_open(%s,%s) failed!\n",naten,ucs4i));
+		}
+	iconv_handle_U2Latin1 = UT_iconv_open ("ISO-8859-1", ucs4i);
+	if (!UT_iconv_isValid(iconv_handle_U2Latin1))
+		{
+			UT_DEBUGMSG(("WARNING: UT_iconv_open(ISO-8859-1,%s) failed!\n",ucs4i));
+		}
 	
 	char* winencname = wvLIDToCodePageConverter(that->getWinLanguageCode());
-	iconv_handle_Win2U = UT_iconv_open("UCS-2",winencname);
-	iconv_handle_U2Win = UT_iconv_open(winencname,"UCS-2");
-};
+	iconv_handle_Win2U = UT_iconv_open(ucs4Internal(),winencname);
+	iconv_handle_U2Win = UT_iconv_open(winencname,ucs4Internal());
+}
 
 
 static UT_UCSChar try_CToU(UT_UCSChar c,UT_iconv_t iconv_handle)
 {
-	 /* 
-	   We don't support multibyte chars yet. wcstombcs should be used. 	   
+	if (!UT_iconv_isValid (iconv_handle)) return 0;
+	UT_iconv_reset (iconv_handle);
+
+	/* We don't support multibyte chars yet. wcstombcs should be used.
+	 * NOTE: if so, then we shouldn't be using this function at all! [TODO: ugh.]
 	 */
-	if (c>255)
-		return 0;			
-	if (!UT_iconv_isValid(iconv_handle))
-		return 0;
-	char ibuf[1],obuf[2];			
-	size_t ibuflen = 1, obuflen=2;
-	const char* iptr = ibuf;
-	char* optr = obuf;
-	ibuf[0]	= (unsigned char)c;	
+	if (c > 255)
+		{
+			UT_DEBUGMSG(("WARNING: character code %lx received, substituting 'E'\n", c));
+			c = 'E'; // was return 0, but that can be dangerous...
+		}
+
+	char ibuf[1];
+	char obuf[4];
+	size_t ibuflen = 1;
+	size_t obuflen = 4;
+	const char * iptr = ibuf;
+	char * optr = obuf;
+
+	ibuf[0]	= static_cast<unsigned char>(c);
+
 	size_t donecnt = UT_iconv(iconv_handle,&iptr,&ibuflen,&optr,&obuflen);			
+
+	UT_UCSChar uval = 0;
+
 	if (donecnt!=(size_t)-1 && ibuflen==0) 
 	{
-		unsigned short uval;
-		unsigned short b0 = (unsigned char)obuf[XAP_EncodingManager::swap_stou];
-		unsigned short b1 = (unsigned char)obuf[!XAP_EncodingManager::swap_stou];
-		uval = (b1<<8) | b0;
-		return uval;
-	} else
-		return  0;
-};
+		if (XAP_EncodingManager::swap_stou)
+			{
+				uval = static_cast<UT_UCSChar>(static_cast<unsigned char>(obuf[3]));
+				uval = static_cast<UT_UCSChar>(static_cast<unsigned char>(obuf[2])) | (uval<<8);
+				uval = static_cast<UT_UCSChar>(static_cast<unsigned char>(obuf[1])) | (uval<<8);
+				uval = static_cast<UT_UCSChar>(static_cast<unsigned char>(obuf[0])) | (uval<<8);
+			}
+		else
+			{
+				uval = static_cast<UT_UCSChar>(static_cast<unsigned char>(obuf[0]));
+				uval = static_cast<UT_UCSChar>(static_cast<unsigned char>(obuf[1])) | (uval<<8);
+				uval = static_cast<UT_UCSChar>(static_cast<unsigned char>(obuf[2])) | (uval<<8);
+				uval = static_cast<UT_UCSChar>(static_cast<unsigned char>(obuf[3])) | (uval<<8);
+			}
+	}
+	return uval;
+}
 
 static UT_UCSChar try_UToC(UT_UCSChar c,UT_iconv_t iconv_handle)
 {
-	if (!UT_iconv_isValid(iconv_handle))
-		return 0;
-	char ibuf[2],obuf[10];			
-	size_t ibuflen = sizeof(ibuf), obuflen=sizeof(obuf);
-	const char* iptr = ibuf;
-	char* optr = obuf;
-	{
-		unsigned char b0 = c & 0xff, b1 = c >>8;
-		ibuf[XAP_EncodingManager::swap_utos] = b0;
-		ibuf[!XAP_EncodingManager::swap_utos] = b1;
-	}
-	size_t donecnt = UT_iconv(iconv_handle,&iptr,&ibuflen,&optr,&obuflen);
-	/* reset state */
-	UT_iconv_reset(iconv_handle);
+	if (!UT_iconv_isValid(iconv_handle)) return 0;
+	UT_iconv_reset (iconv_handle);
+
+	char ibuf[4];
+	char obuf[6];
+	size_t ibuflen = 4;
+	size_t obuflen = 6;
+	const char * iptr = ibuf;
+	char * optr = obuf;
+
+	if (XAP_EncodingManager::swap_utos)
+		{
+			ibuf[0] = static_cast<char>((c      ) & 0xff);
+			ibuf[1] = static_cast<char>((c >>  8) & 0xff);
+			ibuf[2] = static_cast<char>((c >> 16) & 0xff);
+			ibuf[3] = static_cast<char>((c >> 24) & 0xff);
+		}
+	else
+		{
+			ibuf[0] = static_cast<char>((c >> 24) & 0xff);
+			ibuf[1] = static_cast<char>((c >> 16) & 0xff);
+			ibuf[2] = static_cast<char>((c >>  8) & 0xff);
+			ibuf[3] = static_cast<char>((c      ) & 0xff);
+		}
+
+	size_t donecnt = UT_iconv (iconv_handle,&iptr,&ibuflen,&optr,&obuflen);
+
+	UT_UCSChar byte = 0;
+
 	if (donecnt!=(size_t)-1 && ibuflen==0) 
 	{
-		int len = sizeof(obuf) - obuflen;
-		if (len!=1)
-			return 0x1ff;/* tell that singlebyte encoding can't represent it*/
-		else
-			return (unsigned char)*obuf;
-	} else
-		return  0;
-};
+		if (obuflen != 5) // grr... [TODO: ugh.]
+			{
+				UT_DEBUGMSG(("WARNING: character code %lx received, substituting 'E'\n", c));
+				byte = static_cast<UT_UCSChar>(static_cast<unsigned char>('E'));
+			}
+		else byte = static_cast<UT_UCSChar>(static_cast<unsigned char>(obuf[0]));
+	}
+	return byte;
+}
 
 UT_UCSChar XAP_EncodingManager::try_nativeToU(UT_UCSChar c) const
 {
 	return try_CToU(c,iconv_handle_N2U);
-};
+}
 
 UT_UCSChar XAP_EncodingManager::try_UToNative(UT_UCSChar c)  const
 {
 	return try_UToC(c,iconv_handle_U2N);
-};
+}
 
 UT_UCSChar XAP_EncodingManager::try_UToLatin1(UT_UCSChar c)  const
 {
 	return try_UToC(c,iconv_handle_U2Latin1);
-};
+}
 
 UT_UCSChar XAP_EncodingManager::try_WindowsToU(UT_UCSChar c) const 
 { 
@@ -417,7 +497,7 @@ static const char* search_rmap(const _rmap* m,const char* key,bool* is_default =
 	if (is_default)	
 		*is_default = true;
 	return m->value;
-};
+}
 
 static const char* search_rmap_with_opt_suffix(const _rmap* m,const char* key,const char* fallback_key=NULL,const char* fallback_key_final=NULL)
 {
@@ -426,7 +506,7 @@ static const char* search_rmap_with_opt_suffix(const _rmap* m,const char* key,co
 	if (!is_default || !fallback_key)
 		return value;
 	return search_rmap_with_opt_suffix(m,fallback_key,fallback_key_final);
-};
+}
 
 
 struct _map
@@ -445,7 +525,7 @@ static const char* search_map(const _map* m,const char* key,bool* is_default = N
 	if (is_default)
 		*is_default = true;
 	return m->value;
-};
+}
 
 static const char* search_map_with_opt_suffix(const _map* m,const char* key,const char* fallback_key=NULL,const char* fallback_key_final=NULL)
 {
@@ -454,7 +534,7 @@ static const char* search_map_with_opt_suffix(const _map* m,const char* key,cons
 	if (!is_default || !fallback_key)
 		return value;
 	return search_map_with_opt_suffix(m,fallback_key,fallback_key_final);
-};
+}
 
 /* ************************* here begin tables *************************/
 
@@ -549,7 +629,7 @@ static const char* zh_TW_big5_keys[]=
 static const _rmap cjk_word_fontname_mapping_data[]=
 {
     {NULL},
-    {(char*)zh_TW_big5,zh_TW_big5_keys},
+    {reinterpret_cast<const char*>(zh_TW_big5),zh_TW_big5_keys},
     {NULL}
 };
 
@@ -566,16 +646,6 @@ static const _rmap langcode_to_cjk[]=
 };
 
 
-/*
- TODO I'm pretty sure you can't break Korean at any character.
- TODO And what about Japanese Katakana and Hiragana?
-*/
-static const _rmap can_break_words_data[]=
-{
-	{"0"}, /* default value - can't break words at any character. */    
-	{"1",cjk_languages},
-	{NULL}
-};
 
 /*
  This table is useful since some iconv implementations don't know some cpNNNN 
@@ -583,14 +653,77 @@ static const _rmap can_break_words_data[]=
 */
 static const _map MSCodepagename_to_charset_name_map[]=
 {
-/*key, value*/
-    {NULL,NULL},
-	// libiconv also lists "SHIFT_JIS", "SHIFT-JIS", "MS_KANJI", "csShiftJIS"
-	{"CP932","SJIS"},
-    {"CP936","GBK"},
-    {"CP950","BIG5"},  
-	{"CP1361","JOHAB"},
-    {NULL,NULL}
+	/*key,		value*/
+	{ NULL,			NULL },
+
+//	{ "CP0",		?? },					// ANSI code page
+//	{ "CP1",		?? },					// OEM code page
+//	{ "CP2",		?? },					// Macintosh code page
+//	{ "CP3",		?? },					// The current thread's ANSI code page (Win2k)
+
+//	{ "CP42",		?? },					// Symbol code page (Win2k)
+
+//	{ "CP037",		?? },					// EBCDIC
+
+	{ "CP437",		"CP437" },				// MS-DOS United States
+
+//	{ "CP500",		?? },					// EBCDIC "500V1"
+
+	{ "CP708",		"ASMO-708" },			// Arabic (ASMO 708)
+//	{ "CP709",		?? },					// Arabic (ASMO 449+, BCON V4)
+//	{ "CP710",		?? },					// Arabic (Transparent Arabic)
+//	{ "CP720",		?? },					// Arabic (Transparent ASMO)
+
+	{ "CP737",		"CP737" },				// Greek (formerly 437G)
+	{ "CP775",		"CP775" },				// Baltic
+	{ "CP850",		"CP850" },				// MS-DOS Multilingual (Latin I)
+	{ "CP852",		"CP852" },				// MS-DOS Slavic (Latin II)
+	{ "CP855",		"CP855" },				// IBM Cyrillic (primarily Russian)
+	{ "CP857",		"CP857" },				// IBM Turkish
+	{ "CP860",		"CP861" },				// MS-DOS Portuguese
+	{ "CP861",		"CP861" },				// MS-DOS Icelandic
+	{ "CP862",		"CP862" },				// Hebrew
+	{ "CP863",		"CP863" },				// MS-DOS Canadian-French
+	{ "CP864",		"CP864" },				// Arabic
+	{ "CP865",		"CP865" },				// MS-DOS Nordic
+	{ "CP866",		"CP866" },				// MS-DOS Russian
+	{ "CP869",		"CP869" },				// IBM Modern Greek
+	{ "CP874",		"CP874" },				// Thai
+//	{ "CP875",		?? },					// EBCDIC
+
+	{ "CP932",		"SJIS" },				// Japanese
+	{ "CP936",		"GBK" },				// Chinese (PRC, Singapore)
+	{ "CP949",		"CP949" },				// Korean
+	{ "CP950",		"BIG5" },				// Chinese (Taiwan; Hong Kong SAR, PRC)
+
+//	{ "CP1026",		?? },					// EBCDIC
+
+	{ "CP1200",		"UCS-2" },				// Unicode (BMP of ISO 10646)	// ??
+
+	{ "CP1250",		"CP1250" },				// Windows 3.1 Eastern European
+	{ "CP1251",		"CP1251" },				// Windows 3.1 Cyrillic
+	{ "CP1252",		"CP1252" },				// Windows 3.1 US (ANSI)
+	{ "CP1253",		"CP1253" },				// Windows 3.1 Greek
+	{ "CP1254",		"CP1254" },				// Windows 3.1 Turkish
+	{ "CP1255",		"CP1255" },				// Hebrew
+	{ "CP1256",		"CP1256" },				// Arabic
+	{ "CP1257",		"CP1257" },				// Baltic
+
+	{ "CP1361",		"JOHAB" },				// Korean (Johab)
+
+	{ "CP10000",	"MACINTOSH" },			// Macintosh Roman (alias name can be MACROMAN, 
+	                                        //  but iconv from glibc seems to not have clue
+	{ "CP10001",	"SJIS" },				// Macintosh Japanese	// ??
+	{ "CP10006",	"MACGREEK" },			// Macintosh Greek I
+	{ "CP10007",	"MACCYRILLIC" },		// Macintosh Cyrillic
+	{ "CP10029",	"MACCENTRALEUROPE" },	// Macintosh Latin 2
+	{ "CP10079",	"MACICELAND" },			// Macintosh Icelandic
+	{ "CP10081",	"MACTURKISH" },			// Macintosh Turkish
+
+	{ "CP65000",	"UTF-7" },				// Translate using UTF-7 (Win2k, NT 4.0)	// ??
+	{ "CP65001",	"UTF-8" },				// Translate using UTF-8 (Win2k, NT 4.0)	// ??
+
+    { NULL,			NULL }
 };
 
 /*
@@ -621,9 +754,12 @@ static const _map langcode_to_winlangcode[]=
    {"zh_CN.GB2312",	"0x804"},     
    {"zh_CN.GBK",	"0x804"}, 
    {"zh_CN.GB18030",	"0x804"}, 
+   {"zh_CN.UTF-8",	"0x804"}, 
    {"zh_HK.BIG5-HKSCS",	"0x404"},  
+   {"zh_HK.UTF-8",	"0x404"},  
    {"zh_TW.BIG5",	"0x404"},  
-    {NULL}
+   {"zh_TW.UTF-8",	"0x404"},  
+   {NULL}
 };
 
 #undef v
@@ -676,6 +812,7 @@ const XAP_LangInfo XAP_EncodingManager::langinfo[] =
     {{   "Chewa",             "",       "",       "langChewa",                         "92"    }},
     {{   "Chinese",           "zh",     "0x0004", "langTradChinese",                   "19"    }},
     {{   "Chinese",           "zh",     "0x0004", "langSimpChinese",                   "33"    }},
+    {{   "Cornish",	          "kw",     "",       "",                                  ""      }},	// Jordi 19/10/2002 
     {{   "Corsican",          "co",     "",       "",                                  ""      }},
     {{   "Croatian",          "hr",     "0x001a", "langCroatian",                      "18"    }},
     {{   "Czech",             "cs",     "0x0005", "langCzech",                         "38"    }},
@@ -741,6 +878,7 @@ const XAP_LangInfo XAP_EncodingManager::langinfo[] =
     {{   "Manx Gaelic",       "gv",     "",       "langGailck",                        "141"   }},
     {{   "Maori",             "mi",     "",       "",                                  ""      }},
     {{   "Marathi",           "mr",     "",       "langMarathi",                       "66"    }},
+    {{   "Marshallese",       "mh",     "",       "",                                  ""      }},
     {{   "Moldavian",         "mo",     "",       "langMoldavian",                     "53"    }},
     {{   "Mongolian",         "mn",     "",       "langMongolian(Mongolian)",          "57"    }},
     {{   "Mongolian",         "mn",     "",       "langMongolianCyr(Cyrillic)",        "58"    }},
@@ -765,6 +903,7 @@ const XAP_LangInfo XAP_EncodingManager::langinfo[] =
     {{   "Samoan",            "sm",     "",       "",                                  "" 	   }},
     {{   "Sangro",            "sg",     "",       "",                                  "" 	   }},
     {{   "Sanskrit",          "sa",     "",       "langSanskrit",                      "65"    }},
+    {{   "Sardinian",         "sc",     "",       "",        			               ""      }}, // Jordi 19/10/2002 
     {{   "Scots Gaelic",      "gd",     "",       "langGaidhlig",                      "140"   }},
     {{   "Serbian",           "sr",     "0x001a", "langSerbian",                       "42"    }},
     {{   "Serbo-Croatian",    "sh",     "",       "",                                  ""      }},
@@ -810,6 +949,200 @@ const XAP_LangInfo XAP_EncodingManager::langinfo[] =
     {{   NULL,                "",       "",       "",                                  "" 	   }}
 };
 
+/* 
+ * Line Breaking tables 
+ */
+
+
+enum EUniCat {
+	NONATOMIC=0, 
+	ATOMIC=1, 
+	PUNCNOEND=2, 
+	PUNCNOSTART=3, 
+	PUNCFORCE=4,
+	UNKNOWN=5
+};
+// Prototype.
+static enum EUniCat categoriseUniChar(UT_UCS4Char c);
+
+struct SCatRange {
+	UT_UCS4Char start;
+	UT_UCS4Char end;
+	enum EUniCat cat;
+};
+
+/* 
+ * This table categorises all known Unicode characters.
+ * The entries are inclusive ranges which must be in numerical order.
+ *
+ * Defaults should be provided by access functions for unknown characters.
+ *
+ * Description of categories:
+ *
+ * NONATOMIC    - The character does not form an atomic (i.e. non-breakable)
+ *                unit. For example, English characters are of this type, 
+ *                since several are needed to make a word.
+ *
+ * ATOMIC       - The character does form an atomic unit. For example, Chinese
+ *                characters are of this type; although several characters
+ *                combine to make a word, it is legal to break between any two
+ *                Chinese characters.
+ *
+ * PUNCNOEND    - This is punctuation which must not be allow to end a line
+ *                (unless forced by PUNCFORCE); for example, an open bracket.
+ *                In both English and Chinese you cannot break before the
+ *                '(' in "(word)".
+ *
+ * PUNCNOSTART  - Like PUNCNOSTART, but for punctuation which cannot start a
+ *                line, e.g: ) , :
+ *
+ * PUNCFORCE    - Punctuation which it is always possible to break before and
+ *                after regardless of what the surrounding characters are. This
+ *                is for punctuation that acts like a space.
+ * 
+ */
+struct SCatRange UniCharCats[] = {
+	{0x20, 0x20, PUNCFORCE},   // Space
+	{0x21, 0x21, PUNCNOSTART}, // !
+	{0x22, 0x27, NONATOMIC},   // "#$%&'
+	{0x28, 0x28, PUNCNOEND},   // ( 
+	{0x29, 0x29, PUNCNOSTART},  // ) 
+	{0x2a, 0x2b, NONATOMIC},   // *+
+	{0x2c, 0x2e, PUNCNOSTART},  // ,-.
+	{0x2f, 0x2f, NONATOMIC},   // /
+	{0x30, 0x39, NONATOMIC},   // Western numerals.
+	{0x3a, 0x3b, PUNCNOSTART}, // :;
+	{0x3c, 0x3c, PUNCNOEND},   // <
+	{0x3d, 0x3d, NONATOMIC},   // =
+	{0x3e, 0x3f, PUNCNOSTART}, // >?
+	{0x40, 0x40, NONATOMIC},   // @
+	{0x41, 0x5a, NONATOMIC},   // Western A-Z.
+	{0x5b, 0x5b, PUNCNOEND},   // [ 
+	{0x5c, 0x5c, NONATOMIC},   // "\"
+	{0x5d, 0x5d, PUNCNOSTART}, // ]
+	{0x5e, 0x60, NONATOMIC},   // ^_`
+	{0x61, 0x7a, NONATOMIC},   // Western a-z.
+	{0x7b, 0x7b, PUNCNOEND},   // {
+	{0x7c, 0x7c, NONATOMIC},   // "\" 
+	{0x7d, 0x7d, PUNCNOSTART}, // }
+	{0x7e, 0x7e, NONATOMIC},   // ~ 
+
+	/* 
+	  Korean Hangul Jamo 
+	  Korean line breaking is similar to English. However, I'm not
+	  sure if these phonetic elements can occur on their own and, if
+	  they can, what the breaking rules should be.
+	*/
+	{0x1100, 0x11ff, NONATOMIC},     // Korean phonetic elements.
+
+	/* General punctuation */
+	{0x2002, 0x2003, PUNCFORCE},     // en-space, em-space
+	{0x2010, 0x2010, PUNCNOSTART},   // Hyphen
+	{0x2013, 0x2014, PUNCFORCE},     // en-dash, em-dash
+	{0x2018, 0x2018, PUNCNOEND},     // Open single quote
+	{0x2019, 0x2019, PUNCNOSTART},   // Close single quote
+	{0x201c, 0x201c, PUNCNOEND},     // Open double quote
+	{0x201d, 0x201d, PUNCNOSTART},   // Close double quote
+	{0x2020, 0x2021, PUNCNOSTART},   // Long cross, double dagger
+	{0x2026, 0x2026, PUNCNOSTART},   // ...
+	{0x2030, 0x2031, PUNCNOSTART},   // Per mille/10000 signs
+	{0x2032, 0x2034, PUNCNOSTART},   // Single, double, triple primes
+
+	/* CJK Blocks */
+
+	{0x3002, 0x3002, PUNCNOSTART},   //  Ideographic full stop
+	{0x3008, 0x3008, PUNCNOEND},     //  CJK <
+	{0x3009, 0x3009, PUNCNOSTART},   //  CJK >
+	{0x300a, 0x300a, PUNCNOEND},     //  CJK <<
+	{0x300b, 0x300b, PUNCNOSTART},   //  CJK >>
+	{0x300c, 0x300c, PUNCNOEND},     //  CJK quote open
+	{0x300d, 0x300d, PUNCNOSTART},   //  CJK quote close
+	{0x300e, 0x300e, PUNCNOEND},     //  CJK thick quote open
+	{0x300f, 0x300f, PUNCNOSTART},   //  CJK thick quote close
+	{0x3010, 0x3010, PUNCNOEND},     //  CJK |(
+	{0x3011, 0x3011, PUNCNOSTART},   //  CJK )|
+	{0x3014, 0x3014, PUNCNOEND},     //  CJK [
+	{0x3015, 0x3015, PUNCNOSTART},   //  CJK ]
+	{0x3016, 0x3016, PUNCNOEND},     //  CJK [(
+	{0x3017, 0x3017, PUNCNOSTART},   //  CJK )]
+	{0x3018, 0x3018, PUNCNOEND},     //  CJK [[
+	{0x3019, 0x3019, PUNCNOSTART},   //  CJK ]]
+	{0x301a, 0x301a, PUNCNOEND},     //  CJK ]|
+	{0x301b, 0x301b, PUNCNOSTART},   //  CJK |[
+	{0x301d, 0x301d, PUNCNOEND},     //  CJK ``
+	{0x301e, 0x301e, PUNCNOSTART},   //  CJK ''
+
+	/* Hangul Compatibility Jamo */
+	{0x3130, 0x318f, NONATOMIC},
+
+	/* More CJK blocks */
+	{0x3200, 0x32ff, ATOMIC},        // Enclosed CJK Letters and Months
+	{0x3300, 0x33ff, ATOMIC},        // CJK Compatibility
+	{0x3400, 0x34ff, ATOMIC},        // CJK Unified Ideographs Ext. A
+	{0x4e00, 0x9faf, ATOMIC},        // CJK Unified Ideographs 
+
+	/* Hangul Syllabuls */
+	{0xac00, 0xd7af, NONATOMIC},
+	
+	/* Another CJK block */
+	{0xf900,  0xfaff, ATOMIC},       // CJK Compatibility Ideographs
+
+	/* Halfwidth and Fullwidth Forms. */
+	{0xff01, 0xff01, PUNCNOSTART},   // !
+	{0xff02, 0xff02, ATOMIC},        // "
+	{0xff05, 0xff05, PUNCNOSTART},   // %
+	{0xff06, 0xff07, ATOMIC},        // &'
+	{0xff08, 0xff08, PUNCNOEND},     // (
+	{0xff09, 0xff09, PUNCNOSTART},   // )
+	{0xff0a, 0xff0b, ATOMIC},        // *+
+	{0xff0c, 0xff0c, PUNCNOSTART},   // ,
+	{0xff0d, 0xff0d, ATOMIC},        // -
+	{0xff0e, 0xff0e, PUNCNOSTART},   // .
+	{0xff0f, 0xff0f, ATOMIC},        // /
+	{0xff10, 0xff19, ATOMIC},        // Numerals
+	{0xff1a, 0xff1b, PUNCNOSTART},   // :;
+	{0xff1c, 0xff1c, PUNCNOEND},     // <
+	{0xff1d, 0xff1d, ATOMIC},        // =
+	{0xff1e, 0xff1e, PUNCNOSTART},   // >
+	{0xff1f, 0xff1f, PUNCNOSTART},   // ?
+	{0xff20, 0xff20, ATOMIC},        // @
+	{0xff21, 0xff3a, ATOMIC},        // A-Z
+	{0xff3b, 0xff3b, PUNCNOEND},     // [
+	{0xff3c, 0xff3c, ATOMIC},        // "\" 
+	{0xff3d, 0xff3d, PUNCNOSTART},   // ]
+	{0xff3e, 0xff5a, ATOMIC},        // ^_`a-z
+	{0xff5b, 0xff5b, PUNCNOEND},     // {
+	{0xff5c, 0xff5c, ATOMIC},        // |
+	{0xff5d, 0xff5d, PUNCNOSTART},   // }
+	{0xff5e, 0xff5e, ATOMIC},        // ~
+	{0xff61, 0xff61, PUNCNOSTART},   // Ideographic full stop
+	{0xff62, 0xff62, PUNCNOEND},     // Halfwidth left corner bracket
+	{0xff63, 0xff63, PUNCNOSTART},   // Halfwidth right corner bracket
+	{0xff64, 0xff64, PUNCNOSTART},   // Halfwidth ideographic comma
+	{0xffe0, 0xffe0, PUNCNOEND},     // Fullwidth Cent sign
+	{0xffe1, 0xffe1, PUNCNOEND},     // Fullwidth Pound sign
+	{0xffe5, 0xffe5, PUNCNOEND},     // Fullwidth Yen sign
+
+    /* More CJK blocks. */
+	
+	{0x20000, 0x2a6df, ATOMIC},      // CJK Unified Ideographs Ext. B
+	{0x2f800, 0x2fa1f, ATOMIC},      // CJK Compatibility Ideographs Sup.
+	{0,0,ATOMIC}
+};
+
+/*
+ * Boolean rules for whether a line break is allowed between all possible
+ * combinations of two categories.
+ */
+static bool blineBreakRules[] = {
+	// 2nd char:   NONATOMIC, ATOMIC, PUNCNOEND, PUNCNOSTART, PUNCFORCE
+/* 1st char    */
+/* NONATOMIC   */  false,     true,   false,     false,       true,
+/* ATOMIC      */  true,      true,   true,      false,       true,
+/* PUNCNOEND   */  false,     false,  false,     false,       true,
+/* PUNCNOSTART */  true,      true,   true,      false,       true,
+/* PUNCFORCE   */  true,      true,   true,      true,        true}; 
+
 /* ************************* here end tables *************************/
 
 const XAP_LangInfo* XAP_EncodingManager::findLangInfo(const char* key,XAP_LangInfo::fieldidx idx)
@@ -821,12 +1154,11 @@ const XAP_LangInfo* XAP_EncodingManager::findLangInfo(const char* key,XAP_LangIn
 		if (!UT_stricmp(cur->fields[idx],key))
 			return cur;
 	return NULL;
-};
+}
 
-bool XAP_EncodingManager::swap_utos = 0;
-bool XAP_EncodingManager::swap_stou = 0;
+bool XAP_EncodingManager::swap_utos = false;
+bool XAP_EncodingManager::swap_stou = false;
 
-UT_Bijection XAP_EncodingManager::cjk_word_fontname_mapping;
 UT_Bijection XAP_EncodingManager::fontsizes_mapping;
 
 void XAP_EncodingManager::initialize()
@@ -851,11 +1183,23 @@ void XAP_EncodingManager::initialize()
 		"UTF-16LE",			// superset
 		"UTF-16-LE",		// my guess
 		0 };
-	const char ** p;
-	UT_iconv_t iconv_handle;
+
+	// UCS-4 Encoding Names
+	static const char * (szUCS4BENames[]) = {
+		"UCS-4BE",			// preferred
+		"UCS-4-BE",			// older libiconv (??)
+		0 };
+	static const char * (szUCS4LENames[]) = {
+		"UCS-4LE",			// preferred
+		"UCS-4-LE",			// older libiconv (??)
+		0 };
+
+	const char ** p = 0;
+	UT_iconv_t iconv_handle = UT_ICONV_INVALID;
+
 	for (p = szUCS2BENames; *p; ++p)
 	{
-		if ((iconv_handle = UT_iconv_open(*p,*p)) != (UT_iconv_t)-1)
+		if ((iconv_handle = UT_iconv_open(*p,*p)) != UT_ICONV_INVALID)
 		{
 			UT_iconv_close(iconv_handle);
 			UCS2BEName = *p;
@@ -864,7 +1208,7 @@ void XAP_EncodingManager::initialize()
 	}
 	for (p = szUCS2LENames; *p; ++p)
 	{
-		if ((iconv_handle = UT_iconv_open(*p,*p)) != (UT_iconv_t)-1)
+		if ((iconv_handle = UT_iconv_open(*p,*p)) != UT_ICONV_INVALID)
 		{
 			UT_iconv_close(iconv_handle);
 			UCS2LEName = *p;
@@ -880,7 +1224,36 @@ void XAP_EncodingManager::initialize()
 	else
 		UT_DEBUGMSG(("This iconv does not support UCS-2LE!\n"));
 
-	if(!strcmp(enc, "UTF-8") || !strcmp(enc, "UTF8") || !strcmp(enc, "utf-8") || !strcmp(enc, "utf8"))
+	for (p = szUCS4BENames; *p; ++p)
+	{
+		if ((iconv_handle = UT_iconv_open(*p,*p)) != UT_ICONV_INVALID)
+		{
+			UT_iconv_close(iconv_handle);
+			UCS4BEName = *p;
+			break;
+		}
+	}
+	for (p = szUCS4LENames; *p; ++p)
+	{
+		if ((iconv_handle = UT_iconv_open(*p,*p)) != UT_ICONV_INVALID)
+		{
+			UT_iconv_close(iconv_handle);
+			UCS4LEName = *p;
+			break;
+		}
+	}
+	if (UCS4BEName)
+		UT_DEBUGMSG(("This iconv supports UCS-4BE as \"%s\"\n",UCS4BEName));
+	else
+		UT_DEBUGMSG(("This iconv does not support UCS-4BE!\n"));
+	if (UCS4LEName)
+		UT_DEBUGMSG(("This iconv supports UCS-4LE as \"%s\"\n",UCS4LEName));
+	else
+		UT_DEBUGMSG(("This iconv does not support UCS-4LE!\n"));
+
+	if(!UT_stricmp(enc, "UTF-8") || !UT_stricmp(enc, "UTF8")
+		|| !UT_stricmp(enc, "UTF-16") || !UT_stricmp(enc, "UTF16")
+		|| !UT_stricmp(enc, "UCS-2") || !UT_stricmp(enc, "UCS2"))
 		m_bIsUnicodeLocale = true;
 	else
 		m_bIsUnicodeLocale = false;
@@ -921,8 +1294,6 @@ void XAP_EncodingManager::initialize()
 	{	
 	    const char* str = search_rmap_with_opt_suffix(langcode_to_cjk,SEARCH_PARAMS);
 	    is_cjk_ = *str == '1';
-	    str = search_rmap_with_opt_suffix(can_break_words_data,SEARCH_PARAMS);
-	    can_break_words_ = *str == '1';
 	}
 	{
 	    if (cjk_locale()) {
@@ -938,13 +1309,6 @@ void XAP_EncodingManager::initialize()
 		TexPrologue = len ? UT_strdup(buf)  : " ";
 	    };
 	}
-	if (cjk_locale()) {
-	    /* load fontname mapping */
-	    UT_Bijection::pair_data* data = (UT_Bijection::pair_data* )search_rmap_with_opt_suffix(
-		    cjk_word_fontname_mapping_data,SEARCH_PARAMS);
-	    if (data)
-			cjk_word_fontname_mapping.add(data);
-	};
 	{
 	    fontsizes_mapping.clear();
 	    const char** fontsizes = cjk_locale() ? cjk_fontsizes : non_cjk_fontsizes;
@@ -959,7 +1323,7 @@ void XAP_EncodingManager::initialize()
 	
 	init_values(this); /*do this unconditionally! */	
 	{
-	    swap_utos = swap_stou = 0;
+	    swap_utos = swap_stou = false;
 	    swap_utos = UToNative(0x20) != 0x20;
 	    swap_stou = nativeToU(0x20) != 0x20;
 	    
@@ -970,11 +1334,6 @@ void XAP_EncodingManager::initialize()
 
 int XAP_EncodingManager__swap_stou,XAP_EncodingManager__swap_utos;
 
-bool XAP_EncodingManager::can_break_words() const
-{
-    return can_break_words_;
-};
-
 /*
     I'm not sure whether any non-cjk language doesn't make distinction
     between upper and lower case of the letter, but let's be prepared.
@@ -987,7 +1346,7 @@ bool XAP_EncodingManager::is_cjk_letter(UT_UCSChar c) const
     if (!cjk_locale())
 	return 0;
     return (c>0xff);
-};
+}
 
 bool XAP_EncodingManager::noncjk_letters(const UT_UCSChar* str,int len) const
 {
@@ -998,29 +1357,32 @@ bool XAP_EncodingManager::noncjk_letters(const UT_UCSChar* str,int len) const
 	    return 0;
     };
     return 1;
-};
+}
 
 /*
-    This one correlates with can_break_words() very tightly.
-        Under CJK locales it returns 1 for cjk letters. 
-    Under non-CJK locales returns 0.
-*/
-bool XAP_EncodingManager::can_break_at(const UT_UCSChar c) const
+ * Returns true or false depending on whether a break between c[0] and c[1]
+ * is permissible.
+ */
+bool XAP_EncodingManager::canBreakBetween(const UT_UCS4Char c[2]) const
 {
-    if (c == UCS_SPACE 
-	 || c == UCS_MINUS 
-	 || c == UCS_HYPHEN
-	 || c == UCS_EN_DASH
-	 || c == UCS_EM_DASH)
-	return 1;
-    return is_cjk_letter(c);
-};
+	UT_uint8 rule;
+
+	// CJK special case: Can't break between two em dashes.
+	if (c[0] == UCS_EM_DASH && c[1] == UCS_EM_DASH)
+		return false;
+
+	// Find rule number based on character categories.
+	rule = categoriseUniChar(c[0]) * 5 + categoriseUniChar(c[1]);
+
+	// Return corresponding answer.
+	return blineBreakRules[rule];
+}
 
 
 const char* XAP_EncodingManager::getTexPrologue() const
 {
     return TexPrologue;
-};
+}
 
 // Warning:
 // This code forces us to use "GB2312", "BIG5", etc instead
@@ -1035,7 +1397,7 @@ const char* XAP_EncodingManager::charsetFromCodepage(int lid) const
     bool is_default;
     const char* ret = search_map(MSCodepagename_to_charset_name_map,cpname,&is_default);
     return is_default ? cpname : ret;
-};
+}
 
 const char* XAP_EncodingManager::CodepageFromCharset(char *charset) const
 {
@@ -1051,39 +1413,44 @@ const char* XAP_EncodingManager::WindowsCharsetName() const
     bool is_default;
     const char* ret = search_map(MSCodepagename_to_charset_name_map,cpname,&is_default);
     return is_default ? cpname : ret;
-};
+}
 
 UT_uint32  XAP_EncodingManager::getWinLanguageCode() const
 {
 	return WinLanguageCode;
-};
+}
 
 UT_uint32  XAP_EncodingManager::getWinCharsetCode() const
 {
 	return WinCharsetCode;
-};
+}
 
 void 	XAP_EncodingManager::describe()
 {
 	UT_DEBUGMSG(("EncodingManager reports the following:\n"
-		"	NativeEncodingName is %s, LanguageISOName is %s,\n"
-		"   Native8BitEncodingName is %s,\n"
-		"	LanguageISOTerritory is %s,  fallbackchar is '%c'\n"		
+		"	LanguageISOName is %s, LanguageISOTerritory is %s\n"		
+		"	NativeEncodingName is %s, NativeSystemEncodingName is %s,\n"
+		"   Native8BitEncodingName is %s, NativeNonUnicodeEncodingName is %s,\n"
+		"	NativeUnicodeEncodingName is %s,\n"
+		"	fallbackchar is '%c'\n"		
 		"	TexPrologue follows:\n"
 		"---8<--------------\n" 
 			"%s" 
 		"--->8--------------\n"
 		
 		"	WinLanguageCode is 0x%04x, WinCharsetCode is %d\n"
-		"	cjk_locale %d, can_break_words %d, swap_utos %d, swap_stou %d\n"
-		,getNativeEncodingName(),getLanguageISOName(),getNative8BitEncodingName(),
-		getLanguageISOTerritory() ? getLanguageISOTerritory() : "NULL",
-		fallbackChar(1072), getTexPrologue(),getWinLanguageCode(),
-		 getWinCharsetCode(),
-		int(cjk_locale()), int(can_break_words()),int(swap_utos),int(swap_stou)
+		"	cjk_locale %d, swap_utos %d, swap_stou %d\n",
+		getLanguageISOName(), getLanguageISOTerritory() ? getLanguageISOTerritory() : "NULL",
+		getNativeEncodingName(),getNativeSystemEncodingName(),
+		getNative8BitEncodingName(),getNativeNonUnicodeEncodingName(),
+		getNativeUnicodeEncodingName(),
+		fallbackChar(1072),
+		getTexPrologue(),
+		getWinLanguageCode(), getWinCharsetCode(),
+		int(cjk_locale()), int(swap_utos),int(swap_stou)
 		));
 	UT_ASSERT( UT_iconv_isValid(iconv_handle_N2U) && UT_iconv_isValid(iconv_handle_U2N) );
-};
+}
 
 
 /*
@@ -1115,9 +1482,9 @@ const char** localeinfo_combinations(const char* prefix,const char* suffix,const
 		buf[idx++] += suffix;
 	}
 
-    UT_String lang = XAP_EncodingManager::get_instance()->getLanguageISOName();
-	UT_String territory = XAP_EncodingManager::get_instance()->getLanguageISOTerritory();
-	UT_String enc = XAP_EncodingManager::get_instance()->getNativeEncodingName();
+    UT_String lang (XAP_EncodingManager::get_instance()->getLanguageISOName());
+	UT_String territory (XAP_EncodingManager::get_instance()->getLanguageISOTerritory());
+	UT_String enc (XAP_EncodingManager::get_instance()->getNativeEncodingName());
 
 	buf[idx] += sep;
 	buf[idx] += lang;
@@ -1144,11 +1511,9 @@ const char** localeinfo_combinations(const char* prefix,const char* suffix,const
 	for (size_t j = 0; j < 5; ++j)
 		ptrs[j] = buf[j].c_str();
 	ptrs[5] = 0;
-
-	//_DEBUGMSG(("combinations: %s, %s, %s, %s, %s\n", ptrs[0], ptrs[1],ptrs[2],ptrs[3],ptrs[4]));
 	
     return ptrs;
-};
+}
 
 /* pspell hack */
 extern "C" {
@@ -1158,3 +1523,58 @@ const char * xap_encoding_manager_get_language_iso_name(void)
 }
 
 }
+
+
+/*!
+ * Return the line breaking catagory that "c" belongs to.
+ *
+ * Note: For performance reasons this function assumes that the entries
+ *       in UniCharCats are in NUMERICAL ORDER starting with the smallest 
+ *       value.
+ */
+
+static int s_compare_unichar_cats(const void * pC, const void *puc)
+{
+	UT_UCS4Char c = *((UT_UCS4Char*)pC);
+	SCatRange * pUC = (SCatRange*) puc;
+
+	if(c < pUC->start)
+		return -1;
+	else if(c > pUC->end)
+		return 1;
+
+	return 0;
+}
+
+static enum EUniCat categoriseUniChar(UT_UCS4Char c) {
+	enum EUniCat cat=UNKNOWN;
+
+	// use linear search for the bottom (western part of the table, and bsearch for the
+	// rest
+
+	SCatRange * pUC = (SCatRange*)bsearch(&c, UniCharCats, NrElements(UniCharCats), sizeof(SCatRange),
+											  s_compare_unichar_cats);
+
+	if(pUC)
+		cat = pUC->cat;
+
+	/*
+	 * Crude defaults:
+	 *
+	 * If the character is not listed then assume it's nonatomic (like western
+	 * letters) for all code blocks below "Armenian". If it belongs to the
+	 * "Armenian" block or above we assume CJK like atomic letters.
+	 *
+	 * This is not sensible, but it should at least mean that Greek, Cyrillic,
+	 * maybe Korean, Chinese and maybe Japanese get handled OK.
+	 */
+	if (cat == UNKNOWN)
+	{
+		if (c <0x0530) 
+			cat = NONATOMIC;
+		else
+			cat = ATOMIC;
+	}
+	return cat;
+}
+
