@@ -164,7 +164,7 @@ void pt_PieceTable::_unlinkFrag(pf_Frag * pf,
 		*pfragOffsetEnd = 0;
 
 	pf_Frag * pp = pf->getPrev();
-
+	xxx_UT_DEBUGMSG(("Unlink frag %x of type %d \n",pf,pf->getType()));
 	m_fragments.unlinkFrag(pf);
 
 	if (   pp
@@ -188,6 +188,7 @@ void pt_PieceTable::_unlinkFrag(pf_Frag * pf,
 			delete pnt;
 		}
 	}
+	UT_ASSERT(pp->getNext() != pf);
 }
 
 bool pt_PieceTable::_struxHasContent(pf_Frag_Strux * pfs) const
@@ -560,7 +561,57 @@ PT_DocPosition pt_PieceTable::getStruxPosition(PL_StruxDocHandle sdh) const
 
 void pt_PieceTable::deleteHdrFtrStrux(pf_Frag_Strux * pfs)
 {
-	_realDeleteHdrFtrStrux(pfs);
+	UT_return_if_fail( pfs );
+	
+	if(m_pDocument->isMarkRevisions())
+	{
+		const pf_Frag * pfFrag = NULL;
+		pfFrag = static_cast<const pf_Frag *>(pfs);
+		PT_DocPosition dpos1 = getFragPosition(pfFrag);
+
+		pfFrag = pfFrag->getNext();
+
+		while(pfFrag)
+		{
+			if((pfFrag->getType() == pf_Frag::PFT_EndOfDoc) ||
+			   (pfFrag->getType() == pf_Frag::PFT_Strux &&
+				static_cast<const pf_Frag_Strux*>(pfFrag)->getStruxType() == PTX_SectionHdrFtr))
+			{
+				break;
+			}
+
+			pfFrag = pfFrag->getNext();
+		}
+
+		// there should always be at least EndOfDoc fragment !!!
+		UT_return_if_fail( pfFrag );
+		
+		PT_DocPosition dpos2 = getFragPosition(pfFrag);
+
+		UT_uint32 iRealDelete = 0;
+		deleteSpan(dpos1, dpos2, NULL, iRealDelete, true /*delete table struxes*/, false /* don't glob */);
+	}
+	else
+	{
+		const PP_AttrProp * pAP = NULL;
+		UT_return_if_fail(pfs->getStruxType()==PTX_SectionHdrFtr);
+		pf_Frag_Strux_SectionHdrFtr * pfHdr = static_cast<pf_Frag_Strux_SectionHdrFtr *>(pfs);
+
+		if(!getAttrProp(pfHdr->getIndexAP(),&pAP) || !pAP)
+			return;
+
+		const XML_Char * pszHdrId;
+		if(!pAP->getAttribute("id", pszHdrId) || !pszHdrId)
+			return;
+
+		const XML_Char * pszHdrType;
+		if(!pAP->getAttribute("type", pszHdrType) || !pszHdrType)
+			return;
+
+		_realDeleteHdrFtrStrux(pfs);
+		_fixHdrFtrReferences(pszHdrType, pszHdrId);
+	}
+	
 }
 
 void pt_PieceTable::_realDeleteHdrFtrStrux(pf_Frag_Strux * pfs)
