@@ -524,7 +524,7 @@ void s_LaTeX_Listener::_openParagraph(PT_AttrPropIndex api)
 				if (height < 0.9 || height > 1.1)
 				{
 					m_pie->write("\\begin{spacing}{");
-					printf("m_bLineHeight = true\n");
+					UT_DEBUGMSG(("m_bLineHeight = true\n"));
 					m_bLineHeight = true;
 				}
 				if (height > 1.4 && height < 1.6)
